/* Copyright (C) 2002, 2003, 2005 Free Software Foundation, Inc.
   This file is part of the GNU C Library.

   The GNU C Library is free software; you can redistribute it and/or
   modify it under the terms of the GNU Lesser General Public
   License as published by the Free Software Foundation; either
   version 2.1 of the License, or (at your option) any later version.

   The GNU C Library is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   Lesser General Public License for more details.

   You should have received a copy of the GNU Lesser General Public
   License along with the GNU C Library; if not, write to the Free
   Software Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA
   02111-1307 USA.  */

#include <sched.h>
#include <signal.h>
#include <string.h>	/* For the real memset prototype.  */
#include <sysdep.h>
#include <unistd.h>
#include <sys/wait.h>
#include <bits/libc-lock.h>
#include <kernel-features.h>

/* We have to and actually can handle cancelable system().  The big
   problem: we have to kill the child process if necessary.  To do
   this a cleanup handler has to be registered and is has to be able
   to find the PID of the child.  The main problem is to reliable have
   the PID when needed.  It is not necessary for the parent thread to
   return.  It might still be in the kernel when the cancellation
   request comes.  Therefore we have to use the clone() calls ability
   to have the kernel write the PID into the user-level variable.  

   We must use cleanup_id, instead of the existing pid variable, when
   using vfork, so that we do not overwrite the parent's pid variable.
   Otherwise, if the child execs, the parent is cancelled, but vfork
   has not yet returned in the parent, the pid variable will have the
   value zero, not the child's PID.  */
#ifndef FORK
/* Use vfork so that if running without overcommit we do not make a
   copy of the parent's memory.  */
# define FORK_IS_VFORK
# if defined __ASSUME_CLONE_THREAD_FLAGS	
#  define USE_CLEANUP_PID
#  define FORK()						\
  INLINE_SYSCALL (clone, 3,					\
                  CLONE_PARENT_SETTID | CLONE_VFORK | SIGCHLD,	\
                  0, &cleanup_pid)
# else
#  define FORK() __vfork()
# endif
#endif

#ifdef _LIBC_REENTRANT
static void cancel_handler (void *arg);

#ifdef USE_CLEANUP_PID
# define CLEANUP_HANDLER						\
  {									\
    pid_t cleanup_pid;							\
    __libc_cleanup_region_start (1, cancel_handler, &cleanup_pid)
#else
# define CLEANUP_HANDLER					\
  {								\
    __libc_cleanup_region_start (1, cancel_handler, &pid)
#endif

# define CLEANUP_RESET				\
    __libc_cleanup_region_end (0);		\
  }
#endif


/* Linux has waitpid(), so override the generic unix version.  */
#include <sysdeps/posix/system.c>


#ifdef _LIBC_REENTRANT
/* The cancellation handler.  */
static void
cancel_handler (void *arg)
{
  pid_t child = *(pid_t *) arg;

  INTERNAL_SYSCALL_DECL (err);
  INTERNAL_SYSCALL (kill, err, 2, child, SIGKILL);

  TEMP_FAILURE_RETRY (__waitpid (child, NULL, 0));

  DO_LOCK ();

  if (SUB_REF () == 0)
    {
      (void) __sigaction (SIGQUIT, &quit, (struct sigaction *) NULL);
      (void) __sigaction (SIGINT, &intr, (struct sigaction *) NULL);
    }

  DO_UNLOCK ();
}
#endif
